"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetadataManager = void 0;
const typescript_1 = require("typescript");
/**
 * The `MetadataManager` class provides utilities for managing metadata entries
 * within a `@Module` decorator in a TypeScript source file. It allows for the
 * insertion of new metadata entries, merging symbols with static options, and
 * handling various scenarios such as empty module decorators or existing metadata
 * properties.
 */
class MetadataManager {
    constructor(content) {
        this.content = content;
    }
    /**
     * Inserts a new metadata entry into the `@Module` decorator.
     *
     * @param metadata - The metadata key to insert.
     * @param symbol - The symbol to insert.
     * @param staticOptions - The static options to insert.
     * @returns The new content string.
     */
    insert(metadata, symbol, staticOptions) {
        const source = (0, typescript_1.createSourceFile)('filename.ts', this.content, typescript_1.ScriptTarget.ES2017);
        const moduleDecoratorNode = this.findFirstDecoratorMetadata(source, 'Module');
        // If there is no occurrence of `@Module` decorator, nothing will be inserted
        if (!moduleDecoratorNode) {
            return;
        }
        const matchingProperties = moduleDecoratorNode.properties
            .filter((prop) => prop.kind === typescript_1.SyntaxKind.PropertyAssignment)
            .filter((prop) => {
            const name = prop.name;
            switch (name.kind) {
                case typescript_1.SyntaxKind.Identifier:
                    return name.getText(source) === metadata;
                case typescript_1.SyntaxKind.StringLiteral:
                    return name.text === metadata;
                default:
                    return false;
            }
        });
        symbol = this.mergeSymbolAndExpr(symbol, staticOptions);
        const addBlankLinesIfDynamic = () => {
            symbol = staticOptions ? this.addBlankLines(symbol) : symbol;
        };
        if (matchingProperties.length === 0) {
            const expr = moduleDecoratorNode;
            if (expr.properties.length === 0) {
                addBlankLinesIfDynamic();
                return this.insertMetadataToEmptyModuleDecorator(expr, metadata, symbol);
            }
            else {
                addBlankLinesIfDynamic();
                return this.insertNewMetadataToDecorator(expr, source, metadata, symbol);
            }
        }
        else {
            return this.insertSymbolToMetadata(source, matchingProperties, symbol, staticOptions);
        }
    }
    /**
     * Finds the first `@Module` decorator metadata in the source file.
     *
     * @param source - The source file to search.
     * @param identifier - The identifier to match.
     * @returns The first matching `ObjectLiteralExpression` or undefined.
     */
    findFirstDecoratorMetadata(source, identifier) {
        var _a;
        for (const node of this.getSourceNodes(source)) {
            const isDecoratorFactoryNode = node.kind === typescript_1.SyntaxKind.Decorator &&
                node.expression.kind === typescript_1.SyntaxKind.CallExpression;
            if (!isDecoratorFactoryNode)
                continue;
            const expr = node.expression;
            const isExpectedExpression = ((_a = expr.arguments[0]) === null || _a === void 0 ? void 0 : _a.kind) === typescript_1.SyntaxKind.ObjectLiteralExpression;
            if (!isExpectedExpression)
                continue;
            if (expr.expression.kind === typescript_1.SyntaxKind.Identifier) {
                const escapedText = expr.expression.escapedText;
                const isTargetIdentifier = escapedText
                    ? escapedText.toLowerCase() === identifier.toLowerCase()
                    : true;
                if (isTargetIdentifier) {
                    return expr.arguments[0];
                }
            }
        }
    }
    /**
     * Returns an array of all nodes in the source file.
     */
    getSourceNodes(sourceFile) {
        const nodes = [sourceFile];
        const result = [];
        while (nodes.length > 0) {
            const node = nodes.shift();
            if (node) {
                result.push(node);
                if (node.getChildCount(sourceFile) >= 0) {
                    nodes.unshift(...node.getChildren(sourceFile));
                }
            }
        }
        return result;
    }
    /**
     * Inserts a new metadata entry into an empty `@Module` decorator.
     * This method is called when the `@Module` decorator has no properties.
     * It inserts the metadata and symbol into the decorator.
     * @param expr - The `@Module` decorator node.
     * @param metadata - The metadata key to insert.
     * @param symbol - The symbol to insert.
     * @returns The new content string.
     */
    insertMetadataToEmptyModuleDecorator(expr, metadata, symbol) {
        const position = expr.getEnd() - 1;
        const toInsert = `  ${metadata}: [${symbol}]`;
        return this.content.split('').reduce((content, char, index) => {
            if (index === position) {
                return `${content}\n${toInsert}\n${char}`;
            }
            else {
                return `${content}${char}`;
            }
        }, '');
    }
    /**
     * Inserts a new symbol into an existing metadata property in the `@Module` decorator.
     * This method is called when the metadata property already exists in the decorator.
     * It inserts the symbol into the existing metadata property.
     * @param source - The source file.
     * @param matchingProperties - The matching metadata properties.
     * @param symbol - The symbol to insert.
     * @param staticOptions - The static options to insert.
     * @returns The new content string.
     */
    insertNewMetadataToDecorator(expr, source, metadata, symbol) {
        const node = expr.properties[expr.properties.length - 1];
        const position = node.getEnd();
        const text = node.getFullText(source);
        const matches = text.match(/^\r?\n\s*/);
        let toInsert;
        if (matches) {
            toInsert = `,${matches[0]}${metadata}: [${symbol}]`;
        }
        else {
            toInsert = `, ${metadata}: [${symbol}]`;
        }
        return this.content.split('').reduce((content, char, index) => {
            if (index === position) {
                return `${content}${toInsert}${char}`;
            }
            else {
                return `${content}${char}`;
            }
        }, '');
    }
    /**
     * Inserts a new symbol into an existing metadata property in the `@Module` decorator.
     * This method is called when the metadata property already exists in the decorator.
     * It inserts the symbol into the existing metadata property.
     * @param source - The source file.
     * @param matchingProperties - The matching metadata properties.
     * @param symbol - The symbol to insert.
     * @param staticOptions - The static options to insert.
     * @returns The new content string.
     */
    insertSymbolToMetadata(source, matchingProperties, symbol, staticOptions) {
        const assignment = matchingProperties[0];
        let node;
        const arrLiteral = assignment.initializer;
        if (!arrLiteral.elements) {
            // "imports" is not an array but rather function/constant
            return this.content;
        }
        if (arrLiteral.elements.length === 0) {
            node = arrLiteral;
        }
        else {
            node = arrLiteral.elements;
        }
        if (Array.isArray(node)) {
            const nodeArray = node;
            const symbolsArray = nodeArray.map((childNode) => childNode.getText(source));
            if (symbolsArray.includes(symbol)) {
                return this.content;
            }
            node = node[node.length - 1];
        }
        let toInsert;
        let position = node.getEnd();
        if (node.kind === typescript_1.SyntaxKind.ArrayLiteralExpression) {
            position--;
            toInsert = staticOptions ? this.addBlankLines(symbol) : `${symbol}`;
        }
        else {
            const text = node.getFullText(source);
            const itemSeparator = (text.match(/^\r?\n(\r?)\s+/) ||
                text.match(/^\r?\n/) ||
                ' ')[0];
            toInsert = `,${itemSeparator}${symbol}`;
        }
        return this.content.split('').reduce((content, char, index) => {
            if (index === position) {
                return `${content}${toInsert}${char}`;
            }
            else {
                return `${content}${char}`;
            }
        }, '');
    }
    /**
     * Merges a symbol with static options into a single string.
     * @param symbol - The symbol to merge.
     * @param staticOptions - The static options to merge.
     * @returns The merged string.
     */
    mergeSymbolAndExpr(symbol, staticOptions) {
        if (!staticOptions) {
            return symbol;
        }
        const spacing = 6;
        let options = JSON.stringify(staticOptions.value, null, spacing);
        options = options.replace(/"([^(")"]+)":/g, '$1:');
        options = options.replace(/"/g, `'`);
        options = options.slice(0, options.length - 1) + '    }';
        symbol += `.${staticOptions.name}(${options})`;
        return symbol;
    }
    /**
     * Adds blank lines around an expression.
     */
    addBlankLines(expr) {
        return `\n    ${expr}\n  `;
    }
}
exports.MetadataManager = MetadataManager;
