"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ModuleImportDeclarator = void 0;
const core_1 = require("@angular-devkit/core");
const path_solver_1 = require("./path.solver");
/**
 * The `ModuleImportDeclarator` class is responsible for declaring module imports
 * in a TypeScript file. It provides methods to insert import statements into the
 * file content based on the provided declaration options.
 */
class ModuleImportDeclarator {
    constructor(solver = new path_solver_1.PathSolver()) {
        this.solver = solver;
    }
    /**
     * Declares a module import in the given content string.
     *
     * @param content - The content of the file where the import will be declared.
     * @param options - The options for the declaration, including the symbol and path.
     * @returns The updated content string with the import statement inserted.
     */
    declare(content, options) {
        const toInsert = this.buildLineToInsert(options);
        const contentLines = content.split('\n');
        const finalImportIndex = this.findImportsEndpoint(contentLines);
        contentLines.splice(finalImportIndex + 1, 0, toInsert);
        return contentLines.join('\n');
    }
    /**
     * Finds the endpoint of the existing import statements in the content.
     *
     * @param contentLines - The lines of the content.
     * @returns The index of the last import statement.
     */
    findImportsEndpoint(contentLines) {
        const reversedContent = Array.from(contentLines).reverse();
        const reverseImports = reversedContent.filter((line) => line.match(/\} from ('|")/));
        if (reverseImports.length <= 0) {
            return 0;
        }
        return contentLines.indexOf(reverseImports[0]);
    }
    /**
     * Builds the import line to insert based on the declaration options.
     *
     * @param options - The options for the declaration, including the symbol and path.
     * @returns The import line to insert.
     */
    buildLineToInsert(options) {
        return `import { ${options.symbol} } from '${this.computeRelativePath(options)}';`;
    }
    /**
     * Computes the relative path for the import statement.
     *
     * @param options - The options for the declaration, including the symbol and path.
     * @returns The relative path for the import statement.
     */
    computeRelativePath(options) {
        let importModulePath;
        if (options.type !== undefined) {
            importModulePath = (0, core_1.normalize)(`/${options.path}/${options.name}.${options.type}`);
        }
        else {
            importModulePath = (0, core_1.normalize)(`/${options.path}/${options.name}`);
        }
        return this.solver.relative(options.module, importModulePath);
    }
}
exports.ModuleImportDeclarator = ModuleImportDeclarator;
